// Squawk messenger. 
// Copyright (C) 2019  Yury Gubich <blue@macaw.me>
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#ifndef CORE_CACHE_HPP
#define CORE_CACHE_HPP
#include "cache.h"

template <class T>
Core::Cache<T>::Cache(const QString& name):
    storage(name),
    cache(new std::map<QString, T> ()),
    abscent(new std::set<QString> ()) {}

template <class T>
Core::Cache<T>::~Cache() {
    close();
    delete cache;
    delete abscent;
}

template <class T>
void Core::Cache<T>::open() {
    storage.open();}

template <class T>
void Core::Cache<T>::close() {
    storage.close();}

template <class T>
void Core::Cache<T>::addRecord(const QString& key, const T& value) {
    storage.addRecord(key, value);
    cache->insert(std::make_pair(key, value));
    abscent->erase(key);
}

template <class T>
T Core::Cache<T>::getRecord(const QString& key) const {
    typename std::map<QString, T>::const_iterator itr = cache->find(key);
    if (itr == cache->end()) {
        if (abscent->count(key) > 0) {
            throw Archive::NotFound(key, storage.getName().toStdString());
        }

        try {
            T value = storage.getRecord(key);
            itr = cache->insert(std::make_pair(key, value)).first;
        } catch (const Archive::NotFound& error) {
            abscent->insert(key);
            throw error;
        }
    }

    return itr->second;
}

template<class T>
bool Core::Cache<T>::checkRecord(const QString& key) const {
    typename std::map<QString, T>::const_iterator itr = cache->find(key);
    if (itr != cache->end())
        return true;

    if (abscent->count(key) > 0)
        return false;

    try {
        T value = storage.getRecord(key);
        itr = cache->insert(std::make_pair(key, value)).first;
    } catch (const Archive::NotFound& error) {
        return false;
    }

    return true;
}

template<typename T>
void Core::Cache<T>::changeRecord(const QString& key, const T& value) {
    storage.changeRecord(key, value);   //there is a non straightforward behaviour: if there was no element at the sorage it will be added
    cache->at(key) = value;
    abscent->erase(key);                //so... this line here is to make it coherent with the storage
}

template<typename T>
void Core::Cache<T>::removeRecord(const QString& key) {
    storage.removeRecord(key);
    cache->erase(key);
    abscent->insert(key);
}

#endif //CORE_CACHE_HPP
